# *** BL_CSV_1.4 PY - 3 Feb 2020  01:24:31 - JKNAUTH

name_and_version = 'BL_CSV version 1.4'

# BL_CSV produces reports about letter allocation among pollbooks.  See
# the User Guide for details.

# Syntax:  BL_CSV.py  input-file|?  output-file|=  maxpages=nbr  mixrem=yes|no

# There is no default for the input (CSV) file.  The default output file
# is CSVout.txt.  The default for maxpages (maximum number of pages per
# pollbook) is 200.  The default for mixrem is no, i.e., no sharing of a
# split letter's remainder pollbook.

# BL_CSV produces several reports about distributing letters (first
# letter of voter's last name) across precinct pollbooks.  It assumes a
# fixed number of names per pollbook page and a standard maximum number
# of pages in a pollbook (maxpages).  All last names on any page start
# with the same letter.

# Input to BL_CSV is a Comma Separated Values (CSV) file created from a
# WCBOE Excel file; the WCBOE file lists data for a number of precincts.
# For each precinct for each letter of the alphabet, the files list the
# number of voters whose last names start with that letter.  The files
# round up each count per letter to the next multiple of the maximum
# number of names per page and calculate the required number of pages
# given that rounding.  Letters with no assigned voters are optionally
# not listed.

# Output from BL_CSV is a number of reports.  One report lists all the
# precinct letters whose required number of voter pages exceeds an alert
# value (the alert_size variable), which is currently coded as the
# minimum of maxpages and 150.

# Another report shows the result of letter allocation across pollbooks.
# Letters can be put in as few pollbooks as possible, called "packed"
# allocation; many letters are split.  Or they can be allocated in
# various ways to reduce letter splitting, called "unpacked" allocation.
# A "big" letter is one whose required number of pages exceeds maxpages.
# For "unpacked" allocation, only "big" letters are split.  The mixrem
# parameter controls whether other letters can share pollbooks with
# "big" letters.  When there are no "big" letters, the setting of mixrem
# does not matter since there is no "big" letter splitting to be done.

# Final reports summarize various things from all the precinct-specific
# reports.


import sys, os, math, os.path
from datetime import datetime


#=====================  INITIALIZE MAIN VARIABLES  =====================

# $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
# The max_pages_per_book, alert_size, and max_limit variables are set up
# in the PROCESS COMMAND PARAMETERS section near the end of the file.

# Note that the following names_per_page variable does little.  It is
# used to calculate the maximum number of names allowed in any pollbook.
# That number is then printed in the prolog of the output.  The
# names_per_page variable is NOT used to calculate the rounded-up number
# of pages per letter for each letter in a precinct; that rounded-up
# number is pulled from the CSV file and is not calculated by BL_CSV.
names_per_page = 5              # Must be changed if CSV format changes

small_limit = 10                # Maximum number of pages for a pollbook
                                # to be designated a "small pollbook"
#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

#-----------------------------------------------------------------------
# Initialize these variables to prepare for scanning the first precinct.
# They will later be reinitialized in finish_precinct() to prepare for
# processing the next precinct.

old_precinct = 'first'     # A dummy "previous precinct" number
                           #
voter_count = 0            # Running counts for a precinct
ceiling_count = 0          #
page_count = 0             #
                           #
exceeded_alert_count = 0   # Number of letters that exceeded the alert
                           # size in this precinct
                           #
ltr_nbr_freqs = []         # Letter frequencies: list of relative
                           # numbers of voters whose last names start
                           # with a given letter; the entries are in
                           # terms of the number of pages required for
                           # each letter and are in order corresponding
                           # to A-Z , e.g.,
                           #   [25, 88, 57, 0, 20, ...]
                           #
for i in range(26):        # Initialize all frequencies to 0's
   ltr_nbr_freqs.append(0) #
#-----------------------------------------------------------------------


#.......................................................................
# These variables will gather data for all precincts and are not
# reinitialized when a new precinct is encountered.

exceeded_alert_size = []      # List with data about letters whose pages
                              # exceeded the alert size; there is a
                              # separate entry for each such problematic
                              # letter in each precinct -- one letter
                              # per entry and possibly multiple entries
                              # per precinct:
                              #   [['precinct', 'ltr', 'nbr_pages'], ...]
                              #
precincts_with_omax_ltrs = [] # List of precincts with letters whose
                              # pages exceeded the alert size; the entry
                              # also has the count of such letters in
                              # the precinct:
                              #   [['precinct', 'nbr_omax_ltrs'], ...]
                              #
compare = []                  # List of data for "packed" vs. "unpacked":
                              #   [['precinct',
                              #      nbr_ubooks, nbr_pbooks,
                              #      nbr_usplits, nbr_psplits,
                              #      nbr_usmall_books, nbr_psmall_books],
                              #    ...]
#.......................................................................

letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

# Grand totals for all precincts.
gt_voters = 0                 #
gt_ceilings = 0               #
gt_pages = 0                  #
gt_precincts = 0              # Number of precincts


#$$$$$$$$$$$$  MAIN PROCEDURE RESUMES AFTER FUNCTION DEFS  $$$$$$$$$$$$$


#=====================  FUNCTION:  HANDLE LETTER  ======================
#
# This function processes a letter and page-count-per-letter for the
# precinct currently being processed.  The page count is added as a
# frequency to the letter frequency list.  That list matches the order
# of A to Z and was initialized with 0's, so any missing letters in the
# CSV file are defaulted to a 0 frequency.  A check is made to ensure
# the input is not a repeat of a previously entered letter.  Running
# sums are incremented for the precinct's voter_count, ceiling_count,
# and page_count.  Exceptional page-frequency values are logged for a
# later report.

def handle_letter(ltr, pages, precinct):
   global voter_count, ceiling_count, page_count, exceeded_alert_count
   freq_index = ord(ltr) - ord('A')      # Find letter's alpha index
   if ltr_nbr_freqs[freq_index] != 0:    # Check for duplicate letters
      sys.exit('\nERROR =====> Letter "' + ltr + '" was repeated in precinct ' + precinct + '.')
   ltr_nbr_freqs[freq_index:freq_index+1] = [int(pages)]
   if int(pages) > alert_size:           # Check for exceptional letter
      # Collect data for later "Letter Size Alert Report"
      exceeded_alert_size.append([precinct, ltr, pages])  # Log it
      exceeded_alert_count += 1
   voter_count += int(voters)            # Increment running counts
   ceiling_count += int(ceiling)         #
   page_count += int(pages)              #
   return


#====================  FUNCTION:  FINISH PRECINCT  =====================
#
# This function processes the data for a precinct.  It is called when
# all the lines in a precinct section of the input file have been found,
# either because of a precinct change detected during the scan of the
# input file or end-of-file was reached for the input file.  The caller
# has gathered all the data for the precinct and has made it available
# in global variables.

# First write the original (now reformatted) input data to the output
# file.  Call a function to calculate the letter allocations and write
# those results to the output file.  Finally, reset variables to prepare
# for handling the next precinct.

def finish_precinct():
   global ltr_nbr_freqs, line_lastINI
   global old_precinct, line_precinct
   global voter_count, ceiling_count, page_count, exceeded_alert_count
   global gt_voters, gt_ceilings, gt_pages, gt_precincts

#------------  Write Original Precinct Data to Output File  ------------

   inp_v =    'VOTERs: '  + str(voter_count)
   inp_c = '   CEILING: ' + str(ceiling_count)
   inp_p = '   PAGES: '   + str(page_count)
   print('\n\n' + old_precinct + ' ' + '='*16 + ' (' + inp_v + inp_c + inp_p + ') ' + '='*16 + '\n', file=outfile)

   freq_hdr = '      A/N   B/O   C/P   D/Q   E/R   F/S   G/T   H/U   I/V   J/W   K/X   L/Y   M/Z'
   print(freq_hdr, file=outfile)        # Output header line

   text_freqs = []                      # Create list of frequency
   for nbr in ltr_nbr_freqs:            # strings for A to Z
      text_freqs.append('{0:>6}'.format(nbr))

   ATOM = []                            # Output A to M frequency strings
   for i in range(13):                  #
      ATOM.append(text_freqs[i])        #
   sATOM = ''.join(ATOM)                #
   print('   ' + sATOM, file=outfile)   #

   NTOZ = []                            # Output N to Z frequency strings
   for i in range(13,26):               #
      NTOZ.append(text_freqs[i])        #
   sNTOZ = ''.join(NTOZ)                #
   print('   ' + sNTOZ, file=outfile)   #

   print('      ' + '-' * 75, file=outfile)

   print(old_precinct)          # Display precinct to show progress

#-----------------------------------------------------------------------

   process_precinct_data(old_precinct)      # Go calculate allocation of
                                            # letters to pollbooks and
                                            # write the results to the
                                            # output file

   gt_voters += voter_count                 # Increase grand totals
   gt_ceilings += ceiling_count             #
   gt_pages += page_count                   #
   gt_precincts += 1                        #

   if exceeded_alert_count > 0:
      precincts_with_omax_ltrs.append([old_precinct, str(exceeded_alert_count)])

   # Initalize variables to prepare for processing the next precinct.
   old_precinct = line_precinct #
   voter_count = 0              #
   ceiling_count = 0            #
   page_count = 0               #
   exceeded_alert_count = 0     #
   ltr_nbr_freqs = []           #
   for i in range(26):          # Initialize all frequencies to 0's
      ltr_nbr_freqs.append(0)   #
   return


#=================  FUNCTION:  PROCESS PRECINCT DATA  ==================
#
# This function determines how letters are to be allocated to pollbooks
# for a precinct.  It uses the precinct's letter frequency list to get
# the number of pages required for each letter.  A report for the
# precinct is written to the output file.

def process_precinct_data(precinct):
   global voter_count, ceiling_count, page_count

# Handle the case where the initial letters have 0 voters (pages).  Skip
# over those letters in later sections by using the saved start_index.
   start_index = -1
   for i in range(26):
       if ltr_nbr_freqs[i] != 0:
          start_index = i
          break
   if start_index == -1:
       sys.exit('\nERROR =====> All letter frequencies are 0.')


#@@@@@@@@@@@@@@  Calculate Data for Unpacked Allocation  @@@@@@@@@@@@@@@

#..................  Divide Letters among Pollbooks  ...................

   book_nbr = 0            # Book number index (1 less than book label)

   book_page_count = 0     # Running count of pages in a pollbook

   good_splits = 0         # Count of splits that don't share letters

   remainder_sp = False    # True = Possible undesirable mix of letters
                           # may occur in a book because a remainder
                           # overflow book left some space which the
                           # next letter might fit into if it has few
                           # enough pages

   unpacked_splits_count = 0  # Count of overflow books created for a
                              # letter page count exceeding the maximum
                              # size; this does not include the first in
                              # the set of all the books created for the
                              # letter

   letters_in_books = [''] # List of letter strings, one string per
                           # pollbook; each string specifies the letters
                           # allocated to that pollbook, e.g.,
                           #   ['A B', 'B C', 'D E F G', ...]

   book_sizes = []         # List of page counts, one per pollbook; each
                           # page count is the size of that pollbook, e.g.,
                           #   [163, 195, ...]

# Loop thru the alphabet, fitting as many pages in each book as possible
# without exceeding the max_pages_per_book limit.  All names on a page
# have the same initial letter for their last names.  For each book,
# find the book's total page count and the last-name initial letters
# allocated to that book.

# Two algorithms are included for the "unpacked" method.  For both, *NO*
# splitting is done for any letter except for those that exceed maximum
# size, the "big" letters.  For parameter mixrem=no (Boolean value is
# False), no split letters share a book with any other letter.  For
# parameter mixrem=yes (Boolean value is True), if a "big" letter is
# split, other letters (all must be unsplit) can share the final book
# (the remainder book) split from the "big" letter; that is the only
# sharing allowed.

# Allocate letters to books, splitting books if necessary so none
# exceeds the maximum page count.  For each letter one of the following
# applies.  The checks are done in this order:

# 1) The letter's pages can fit in the current book (this includes a
#    letter with 0 pages); no new book must be created.

# 2) The letter's pages cannot fit in the current book, but the pages do
#    not exceed the maximum_pages_per_book; then a new book must be
#    created and all the letter's pages will fit in that new book.

# 3) The pages exceed max_pages_per_book; then at least two new books
#    must be created and the pages must be split among them, with full
#    books at the start of the set, each holding the maximum size, and a
#    possible remainder book holding what's left.  The final book might
#    be exactly the maximum size; for that the letter's page count was
#    an exact multiple of the max_pages_per_book and there will be no
#    remainder book.

# For mixrem = True vs. False, see the earlier comment about possibly
# allowing a mix of letters in the remainder book.

   for i in range(start_index,26):
      new_ltr = letters[i]                  # Get the letter in question
      new_ltr_freq = ltr_nbr_freqs[i]       # Get its page count

      test_count = book_page_count + new_ltr_freq  # Check for overflow
      if test_count <= max_pages_per_book:         # if letter is added

# 1) Pages for letter will fit in current book -- no overflow
         book_page_count += new_ltr_freq   # Increment and save count
         book_sizes[book_nbr:book_nbr+1] = [book_page_count]
         letters_so_far = letters_in_books[book_nbr:book_nbr+1][0]
         if new_ltr_freq == 0:
            letters_so_far += ' ' + '-'  # Designate the  missing letter
         elif remainder_sp:   # Check if putting in remainder book space
                              # Flag with "@" to highlight the split mix
            letters_so_far = '@' + letters_so_far[1:] + ' ' + new_ltr
            unpacked_splits_count += 1  # Count the number of bad splits
            remainder_sp = False        # Remainder space not involved
         else:
            letters_so_far += ' ' + new_ltr   # Just add a normal letter
         letters_in_books[book_nbr:book_nbr+1] = [letters_so_far]
#.......................................................................
      elif new_ltr_freq <= max_pages_per_book:
# 2) Pages would overflow the current book, but are not over the book
#    maximum size.  All the letter's pages will fit in the next book, a
#    new book.  Initialize that next book with the current letter as the
#    first letter and the initial book size being the letter's page
#    count.

         remainder_sp = False             # Remainder space not involved
         book_nbr += 1
         book_page_count = new_ltr_freq
         book_sizes.append(new_ltr_freq)
         letters_in_books.append(' ' + new_ltr)
#.......................................................................
      else:
# 3) Letter's page count exceeds the maximum size; must split the pages
#    into a set of two or more new books.

         first_book_for_this_big_ltr = True
         test_count = new_ltr_freq # Use full number of pages to start
         remainder_sp = False      # Initially assume no remainder space

         # Split off the largest possible number of full books to start
         # with; there must be at least one.  That will leave a
         # remainder book, unless the page count is an exact multiple of
         # max_pages_per_book.
         while test_count > max_pages_per_book:

            if i > start_index or not first_book_for_this_big_ltr:
               book_nbr += 1     # Handle the case of the first group of
                                 # letter pages, e.g., for "A", being
                                 # greater than max_pages_per_book.  For
                                 # that case use the first book; don't
                                 # create a new book by incrementing the
                                 # book index for the first book of the
                                 # "A" set.  However the index must be
                                 # incremented for every subsequent book
                                 # in the "A" set

            good_splits += 1     # For mixrem=no, a non-sharing split
            test_count -= max_pages_per_book      # Decrement used pages
            book_sizes.append(max_pages_per_book) # Record current size

            if first_book_for_this_big_ltr:    # Tailor explanation text
               first_book_for_this_big_ltr = False
               add_text = ' ' + new_ltr + ' *** EXCEEDED '+ str(max_pages_per_book) +' ***'
            else:
               add_text = ' ' + new_ltr + '     overflow  '
            letters_in_books[book_nbr:book_nbr+1] = [add_text]

         # At this point, the number of pages left is no longer greater
         # than max_pages_per_book.  It is 0 < pages <= max_pages_per_book.
         # Create the final book with the remaining pages.
         book_nbr += 1
         book_page_count = test_count
         book_sizes.append(test_count)  # Save count of what's left
         letters_in_books.append(' ' + new_ltr + '     overflow  ')

         # mixrem=no forces start of a new empty book.  This prevents
         # any following letter from using any space left in a
         # possible remainder book.  No letter split mixing allowed.
         if mixrem == False:
            book_nbr += 1
            book_page_count = 0
            book_sizes.append(0)
            letters_in_books.append('')
         else:
            remainder_sp = True

# Now go back up to the "for i" to start processing the next letter.


#......................  Delete Empty Pollbooks  .......................

# An empty pollbook can occur when there is a pollbook overflow and all
# the rest of the letters to the end of the alphabet have a 0 frequency
# (no pages for those letters).  Empty books can cause problems later,
# so they are removed here, along with the matching letters_in_books
# string.

   new_sizes = []
   new_letters = []
   for j in range(len(book_sizes)):
      if book_sizes[j] != 0:
         new_sizes.append(book_sizes[j])
         new_letters.append(letters_in_books[j])
   book_sizes = new_sizes
   letters_in_books = new_letters

#..................  Get Summary Counts for Reports  ...................

   unpacked_book_count = len(book_sizes)

   unpacked_splits_count = unpacked_splits_count     # Just documenting
                                                     # for summary
   if mixrem == False: unpacked_splits_count = good_splits

   usmall_books_count = 0
   for k in book_sizes:
      if k <= small_limit: usmall_books_count += 1

#@@@@@@@@@@@@@@@@@@@@  End of Unpacked Processing  @@@@@@@@@@@@@@@@@@@@@


#@@@@@@@@@@@@@@@  Calculate Data for Packed Allocation  @@@@@@@@@@@@@@@@

# Loop thru the alphabet, fitting as many pages in each book as possible
# without exceeding the max_pages_per_book limit.  If an overflow
# occurs, put the remainder into the next book and resume adding other
# letters to that next book.  For a later report, record the letters
# associated with each pollbook, flagging letters that were split.

   pbook_nbr = 0            # Book number index (1 less than book label)
   last_size = max_pages_per_book # Last non-zero book size (page count)
   psplit_count = 0               # Count of letter splits
   current_page_count = 0         # Running count of pages in a pollbook

   ltr_log = []           # Holds paired letter-to-pollbook associations
                          #   [[nbr_book, ' ltr'], ...]

   packed_ltrs = []       # Holds strings of letters per book in order
                          #   ['ltrs_in_book_0', 'ltrs_in_book_1' ...]

   splitting = False              # True = splitting in progress

   for i in range(start_index,26):             # Skip over initial 0's
      ltr = letters[i]                         # Get letter character
      if ltr_nbr_freqs[i] == 0: ltr = '-'      # Flag missing letters
      next_set = ltr_nbr_freqs[i]        # Next set of pages to allocate
      test_count = current_page_count + next_set

      if test_count <= max_pages_per_book:     # Check for overflow
         current_page_count += next_set        # Add the set, which fits
         mltr = ' ' + ltr                      # Usual type of no-split
         ltr_log.append([pbook_nbr, mltr])     # PB# and letter no-split

         if current_page_count != 0:           # Record last non-0 size
            last_size = current_page_count     # used by this branch

      else:
         pbook_nbr += 1    # Current book will fill after add, start new
         needed_to_fill = max_pages_per_book - current_page_count
         if needed_to_fill > 0:
            pb = pbook_nbr - 1             # For log, go back 1 pollbook
            mltr = ' ' + ltr               # Show letter being split
            ltr_log.append([pb, mltr])     # Old PB# and ltr split-start
            psplit_count += 1              # Letter must split
            splitting = True
         else:
            mltr = '!' + ltr                    # Happenstance non-split
            ltr_log.append([pbook_nbr, mltr])   # Record non-split
         next_set -= needed_to_fill             # Adjust for what's left
         if splitting:
            if i == start_index:                # Special handling for
                mltr = '*'+ ltr                 #    first non-0 letter
            else:
               mltr = '>'+ ltr                  # Usual type of split
            ltr_log.append([pbook_nbr, mltr])   # PB# and split-end
            splitting = False
         while next_set > max_pages_per_book:   # Handle huge page count
            pbook_nbr += 1         # Would overflow, so start a new book
            mltr = '*' + ltr                    # "Big" letter non-split
            ltr_log.append([pbook_nbr, mltr])   # Record non-split
            next_set -= max_pages_per_book      # Initialize book count

         if next_set == 0:                      # Record last non-0 size
            last_size = max_pages_per_book      # used by this branch
         else:                                  #
            last_size = next_set                #

         current_page_count = next_set     # Start book with what's left

# Now go back up to the "for i" to start processing the next letter.
#.......................................................................


   if last_size == 0:                      # Just in case
      last_size = max_pages_per_book

   for i in range(pbook_nbr + 1):          # Initialize book lines
      packed_ltrs.append('')               #   ['', '', '', ...]

   # Scan thru the items recorded in ltr_log list, which has this
   # format:  [[nbr_book, ' ltr'], ...]  In the packed_ltrs list,
   # add each letter to the letters accumulated for the appropriate
   # book_nbr.  The result is a packed_ltrs list of letter strings of
   # the accumulated letters per book in book number order, e.g.,
   # [' ltr ltr', ' ltr ltr ltr', ' ltr', ...].  Each string might have
   # a flag character instead of a space for the initial character.

   for item in ltr_log:
      book_ndx = item[0]
      ltr = item[1]
      ltr_string = packed_ltrs[book_ndx] + ltr
      packed_ltrs[book_ndx:book_ndx+1] = [ltr_string]

   # For each book in the precinct, gather data for the book's print
   # line.  For packed books only the last book can have a size other
   # than max_pages_per_book.
   print_pdata = []
   for i in range(pbook_nbr + 1):
      if i == pbook_nbr:                # The final book in the precinct
         pbsize = last_size             # gets a possible non-max size
      else:
         pbsize = max_pages_per_book    # All others get the max size

      # For output, add 1 to the pbook_nbr since book number labels
      # start at 1, not 0.
      print_pdata.append([i+1, pbsize, packed_ltrs[i]])

   # Determine if the book size is small.
   if last_size <= small_limit:
      psmall_books_count = 1
   else:
      psmall_books_count = 0

# The following two commented-out code lines give an approximate count
# of packed splits (right about 99% of the time) because they assume a
# letter split occurs for EVERY new book.  They were replaced by the
# logic above to give an exact count, handling the case where by chance
# a new book begins exactly with a new letter.

#  packed_book_count = math.ceil(page_count/max_pages_per_book)
#  packed_splits_count = packed_book_count - 1    # Close approximation

#@@@@@@@@@@@@@@@@@@@@@  End of Packed Processing  @@@@@@@@@@@@@@@@@@@@@@

#-----  Create and Write Allocation Reports for Current Precinct  ------

   packed_book_count = pbook_nbr + 1       # Also for naming consistency
   packed_splits_count = psplit_count      # in the following code

   inp_ub  =    ' BOOKS: '       + str(unpacked_book_count)    + ' vs. ' + str(packed_book_count)
   inp_us  = '    SPLITS: '      + str(unpacked_splits_count)  + ' vs. ' + str(packed_splits_count)
   inp_usb = '    SMALL BOOKS: ' + str(usmall_books_count)     + ' vs. ' + str(psmall_books_count)
   print('\nUnpacked vs. Packed:   ' + inp_ub + inp_us + inp_usb +'\n', file=outfile)


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   print('USING UNPACKED METHOD', file=outfile)
   print('----- -------- ------\n', file=outfile)

   for i in range(unpacked_book_count):              # Create data lines
      x = '      Book {0:>2d} has {1:>3d} pages and contains letters: {2:s}'
      report_line = x.format(i+1, book_sizes[i], letters_in_books[i])
      print(report_line, file=outfile)               # Write to file

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   print('\nUSING PACKED METHOD', file=outfile)
   print('----- ------ ------\n', file=outfile)

   # Format of print_pdata list:
   #        [[nbr_book_i_label, nbr_book_i_size, 'ltrs_in_book_i'], ...]
   for i in range(packed_book_count):                # Create data lines
      x = '      Book {0:>2d} has {1:>3d} pages and contains letters: {2:s}'
      report_line = x.format(print_pdata[i][0], print_pdata[i][1], print_pdata[i][2])
      print(report_line, file=outfile)               # Write to file

   print('_' * 83, file=outfile)

   # Collect data for later "Unpacked vs. Packed Report"
   compare.append([precinct, unpacked_book_count,   packed_book_count,
                             unpacked_splits_count, packed_splits_count,
                             usmall_books_count,    psmall_books_count])
   return


#=============  FUNCTION:  APPEND MULTI-PRECINCT REPORTS  ==============

# The previous reports written to file were for individual precincts.
# This function outputs reports that consolidate data for all the
# precincts.

def append_multi_reports():
   global gt_voters, gt_ceilings, gt_pages

   print('\n\n=====================  UNPACKED vs. PACKED REPORT  =====================\n', file=outfile)

   print('See the BL_CSV.py User Guide for definitions of the terms used here.\n',  file=outfile)

   # Create a table of the "compare" list data produced by the process_precinct_data function.
   print('Precinct      Unpacked Packed       Unpacked Packed            Unpacked Packed', file=outfile)
   print('--------      -------- ------       -------- ------            -------- ------', file=outfile)

   ubook_sum = 0                  # Initialize sums for the table columns
   pbook_sum = 0                  #
   usplit_sum = 0                 #
   psplit_sum = 0                 #
   usmall_books_sum = 0           #
   psmall_books_sum = 0           #

   for comp in compare:
      ubook_sum += comp[1]        # Increment the column sums
      pbook_sum += comp[2]        #
      usplit_sum += comp[3]       #
      psplit_sum += comp[4]       #
      usmall_books_sum += comp[5] #
      psmall_books_sum += comp[6] #

      comp_data1 = ' {0:<6s}  Books:  {1:>2} vs. {2:>2}'
      comp_data2 = '     Splits: {3:>2} vs. {4:>2}     Small Books: {5:>2} vs. {6:>2}   {0:<6s}'
      comp_data  = (comp_data1 + comp_data2).format(comp[0], comp[1], comp[2], comp[3], comp[4], comp[5], comp[6])
      print(comp_data, file=outfile)

   print('               ----   ----           ----   ----                  --     --', file=outfile)
   print('               {0:>4}   {1:>4}           {2:>4}   {3:>4}                 {4:>3}    {5:>3}'.format(
             ubook_sum, pbook_sum, usplit_sum, psplit_sum, usmall_books_sum, psmall_books_sum), file=outfile)

   print('--------    -------- ------       -------- ------             -------- ------', file=outfile)
   print('Precinct    Unpacked Packed       Unpacked Packed             Unpacked Packed', file=outfile)
   print('                                                                             ', file=outfile)
   if mixrem:
      print('                                     BAD     BAD', file=outfile)
   else:
      print('                                    GOOD     BAD', file=outfile)
   print('                                   SPLITS  SPLITS', file=outfile)


   # Output grand totals for precinct data.
   gt_fmt ='\n\nPrecincts: {0:>3,}    Voters: {1:>7,}    Ceilings: {2:>7,}    Pages: {3:>7,}'
   print(gt_fmt.format(gt_precincts, gt_voters, gt_ceilings, gt_pages), file=outfile)


   print('\n\n======================= LETTER SIZE ALERT REPORT =======================\n', file=outfile)

   # Format of exceeded_alert_size list:  [['precinct', 'ltr', 'nbr_pages'], ...]

   # Skip this report if there was no letter over the alert size for ANY precinct.
   if len(exceeded_alert_size) == 0:
      print('No letters exceeded the alert size' + str(alert_size), file=outfile)

   else:
      print('Letters requiring over ' + str(alert_size) + ' pages in at least one precinct:\n', file=outfile)
      print('               Sorted high to low, count: ' + str(len(exceeded_alert_size)) + '\n', file=outfile)

      # If a letter was over the alert size in a precinct, the data is in the exceeded_alert_size list.
      # Sort on page count (high to low) all letters that were over the alert size in any precinct.
      sorted_max = sorted(exceeded_alert_size, key=lambda x: x[2], reverse=True)
      for item in sorted_max:
         exceeded_max_report = '          Precinct: {0:<6} Letter: {1}  Pages: {2:>4}'.format(item[0], item[1], item[2])
         print(exceeded_max_report, file=outfile)

      print('\n\nPrecincts where at least one letter exceeded ' + str(alert_size) + ' pages: \n', file=outfile)
      print('              Sorted high to low, count: ' + str(len(precincts_with_omax_ltrs)) + '\n', file=outfile)

      # Create a report of all precincts having at least one letter over the alert size, sorted (high to low) on
      # the number of such letters in the precinct.
      sorted_max = sorted(precincts_with_omax_ltrs, key=lambda x: x[1], reverse=True)
      for item in sorted_max:
                                      # item format:  ['precinct', 'nbr_omax_ltrs']
         exceeded_max_report = '          Precinct: {0:<6}  Exceeded Count: {1:>4}'.format(item[0], item[1])
         print(exceeded_max_report, file=outfile)


      print('\n\n==================== EXCEPTIONAL LETTERS REPORT ========================\n', file=outfile)

      # Create a detailed report of precincts with exceptional letters.  This function just
      # consolidates the data.  It calls output_precinct_exceptions to process the data
      # and output the report.

      print('Format:  precinct  (nbr-of-letters, min-pages, max-pages)  letter: pages ...\n ',
             file=outfile)

      # Initialize variables for the first (old) precinct.
      collect_pairs = []   # pairs of problematic letters and the number of
                           # pages each requires, combined as a string
                           #   ['ltr:  nbr_pages', ...]

      pages_list = []      # above pages in the form of integers
                           #   [nbr_pages, nbr_pages, ...]

      fmt_oflow = '{0}: {1:5>}  '

      # format of each entry (itself a list) in the exceeded_alert_size list:
      #    ['precinct', 'ltr', 'nbr_pages']

      old_precinct = exceeded_alert_size[0][0]  # Get first precinct in list

      for entry in exceeded_alert_size:
         if entry[0] == old_precinct:         # Still in old precinct
            # Add to lists the data for this letter in this (old) precinct.
            collect_pairs.append(fmt_oflow.format(entry[1], entry[2]))
            pages_list.append(int(entry[2]))
         else:
            # Go write the report section for the previous precinct.
            output_precinct_exceptions(old_precinct, pages_list, collect_pairs)

            collect_pairs = []          # Initialize variables for next
            pages_list = []             # precinct, the one just hit
            old_precinct = entry[0]     #

            # Add to lists the data for this letter in this (new) precinct
            collect_pairs.append(fmt_oflow.format(entry[1], entry[2]))
            pages_list.append(int(entry[2]))

      # Go write the report section for the final precinct.
      output_precinct_exceptions(old_precinct, pages_list, collect_pairs)
   return


#===============  FUNCTION:  OUTPUT PRECINCT EXCEPTIONS  ===============
#
# Write the section of the Precinct Exceptions Report for the specified
# precinct listed in the exceeded_alert_size list.  Note any letters
# getting close to the maximum pollbook size.

def output_precinct_exceptions(old_precinct, pages_list, collect_pairs):
   minp = min(pages_list)
   minps = str(minp)

   maxp = max(pages_list)
   maxps = str(maxp)
   if maxp > max_limit:
      alert_max = '================> Some are near or over maxpages (' + str(max_pages_per_book) + ')'
   else:
      alert_max = ''

   cntps = str(len(pages_list))
   problem_ltrs = ''.join(collect_pairs)     # List of "letter: size" pairs
   fmt_prt = 'Precinct: {0:<6} ({1:>2},{2:>3},{3:>3})  {4}'
   print(fmt_prt.format(old_precinct, cntps, minps, maxps, problem_ltrs), file=outfile)
   if alert_max != '': print('Precinct: ' + old_precinct + ' ' + alert_max, file=outfile)
   print(' ', file=outfile)
   return


#$$$$$$$$$$$$$$$$$$$$-------------------------------$$$$$$$$$$$$$$$$$$$$
#$$$$$$$$$$$$$$$$$$$$  MAIN PROCEDURE RESUMES HERE  $$$$$$$$$$$$$$$$$$$$
#$$$$$$$$$$$$$$$$$$$$-------------------------------$$$$$$$$$$$$$$$$$$$$

#====================  PROCESS COMMAND PARAMETERS  =====================

#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
max_pages_default = 200                 # Default for max_pages_per_book
#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

if len(sys.argv) < 2 or sys.argv[1] == '?':

   print('\nSyntax:  BL_CSV.py  input-file|?  output-file|=  maxpages=nbr  mixrem=yes|no\n')

   print('There is no default for the input file.  The file can have any name,')
   print('e.g., election35.txt, but it must be in CSV format.\n')

   print('The default output file is CSVout.txt.  The specified or defaulted')
   print('output file will be replaced if it exists.\n')

   print('If "=" is specified for the output file, "OUT_" will be prefixed to the')
   print('input file name to create the output fileid.  If the input file')
   print('extension is ".csv" (case is ignored), the output file extension will be')
   print('changed from ".csv" to ".txt".  For example, an input file named abc.csv')
   print('will result in an output file OUT_abc.txt, but abc.xyz will result in')
   print('OUT_abc.xyz.  An input file of C:\\books\\test.csv will produce an output')
   print('file C:\\books\\OUT_test.txt.\n')

   print('maxpages=nbr controls the maximum number of pages per pollbook.  The')
   print('default for maxpages is '+ str(max_pages_default) + '.\n')

   print('With the "unpacked" letter allocation method, when a "big" letter (one')
   print('with more than the maximum allowed pages per book) is split, it results')
   print('in one or more books of the maximum size and a possible remainder book')
   print("containing the rest of the letter's pages.\n")

   print('mixrem=yes|no controls how strongly the "unpacked" method prevents')
   print('sharing the remainder book between the remaining pages of the split')
   print('"big" letter and pages for any other letter.  Such sharing is a mixing')
   print('of letters in the remainder book.  If the number of pages for a "big"')
   print('letter is an exact multiple of the maximum pages per book, there is no')
   print('remainder book to consider.\n')

   print('mixrem=no, the default, completely prevents such mixing.  Only pages')
   print('from the "big" letter are in its remainder book.\n')

   print('mixrem=yes allows multiple letters to share the remainder book of a')
   print('split "big" letter, but with restrictions.  Only the whole set of pages')
   print('required for another letter (or letters) is allowed to share the')
   print('remainder book with the split "big" letter.  No letters other than the')
   print('"big" letter can have only part of its pages in the shared book; the')
   print('other letters must have all their pages in the book, i.e., not be split')
   print('across any books.')

   sys.exit() # End of Help display


mixrem = False  # "False" prevents any other letter from sharing a "big"
                # letter's remainder book; whereas "True" allows sharing
                # the letter's remainder book.

maxp = max_pages_default                       # Initialize with default

i = 1                   # Scan command line parameters; skip sys.argv[0]
while i < len(sys.argv):
   parm = sys.argv[i].lower()
   if parm == 'mixrem=no':  mixrem = False
   if parm == 'mixrem=yes': mixrem = True
   if parm == 'mixrem':
      sys.exit('\nERROR =====> Parameter must be mixrem=yes or mixrem=no' +
               '  (Use the "?" parameter to see the command syntax.)')
   if (parm.find('=') != -1 and parm != '=' and parm != 'mixrem=no' and parm != 'mixrem=yes'
         and parm[0:9] != 'maxpages='):
      sys.exit('\nERROR =====> Invalid parameter: ' + sys.argv[i] +
               '  (Use the "?" parameter to see the command syntax.)')
   if parm[0:9] == 'maxpages=':
      maxp = parm[9:]
      if maxp.isdigit() == False or int(maxp) < 1:
         sys.exit('\nERROR =====> ' + parm + ' must provide a non-negative integer.')
   i += 1

#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
# Note that max_pages_default was defined at the top of this PROCESS
# COMMAND PARAMETERS section.
max_pages_per_book = int(maxp)   # If it was specified, use the maxpages
                                 # command line parameter; otherwise use
                                 # its default (max_pages_default, which
                                 # was defined above)

# Note that names_per_page was defined in INITIALIZE MAIN VARIABLES near
# the top of this file.
max_names_per_book = names_per_page * max_pages_per_book

# Report a letter if the letter requires more than alert_size pages for
# any precinct
alert_size = min(150, max_pages_per_book)

# Limit used in the Exceptional Letter report. The hardcoded "30" can
# be changed to move the report limit closer to or further from the
# max_pages_per_book (maxpages) value.
max_limit = max_pages_per_book - 30
#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

if len(sys.argv) > 5:
   sys.exit('\nERROR =====> Too many parameters specified.' +
            '  (Use the "?" parameter to see the command syntax.)')

if len(sys.argv) > 3 and sys.argv[3][0:7] != 'mixrem=' and sys.argv[3][0:9] != 'maxpages=':
   sys.exit('\nERROR =====> Invalid parameter in third position: ' + sys.argv[3] +
            '  (Use the "?" parameter to see the command syntax.)')

if len(sys.argv) > 4 and sys.argv[4][0:7] != 'mixrem=' and sys.argv[4][0:9] != 'maxpages=':
   sys.exit('\nERROR =====> Invalid parameter in fourth position: ' + sys.argv[4] +
            '  (Use the "?" parameter to see the command syntax.)')

if len(sys.argv) > 1 and sys.argv[1].find('=') != -1:
   sys.exit('\nERROR ====> A CSV-formatted input file must be specified.' +
            '  A .csv file name extension is not required.')


#====================  GET INPUT AND OUTPUT FILES  =====================

in_fileid = sys.argv[1]

if not os.path.exists(in_fileid):      # Error if cannot find input file
   sys.exit('\nERROR =====> Could not find input file: ' + in_fileid)

if (len(sys.argv) > 2 and
      (sys.argv[2].lower()[0:7] == 'mixrem=' or
       sys.argv[2].lower()[0:9] == 'maxpages=')):
   sys.argv[2] = ''

if len(sys.argv) < 3 or sys.argv[2] == '':
   out_fileid = 'CSVout.txt'
elif sys.argv[2] == '=':
   path_parts = os.path.split(in_fileid)
   drive_dirs = path_parts[0]
   if drive_dirs != '': drive_dirs += '\\'
   file_name = path_parts[1]
   file_name_parts = os.path.splitext(file_name)
   if file_name_parts[1].upper() == '.CSV':
      out_fileid = drive_dirs + 'OUT_' + file_name_parts[0] + '.txt'
   else:
      out_fileid = drive_dirs + 'OUT_' + file_name
else:
   out_fileid = sys.argv[2]

# Windows will catch any unusual file opening problems and report them.
# while aborting the program.
infile  = open(in_fileid, 'r')
outfile = open(out_fileid, 'w')


#====================  CREATE OUTPUT FILE PROLOGUE  ====================

now = datetime.now()
dt_string = now.strftime("%m/%d/%Y   %H:%M:%S")
print('Date and Time:  ' + dt_string + '         ' + name_and_version + '\n', file=outfile)

if mixrem == True: mixrem_text = 'MIXREM=YES'
else:              mixrem_text = 'MIXREM=NO'
print('Input File: ' + in_fileid +  '     Output File: ' + out_fileid + '      ' +
          mixrem_text + '\n', file=outfile)

print('Maximum Pages per Book: ' + str(max_pages_per_book) + '      ' +
      'Names per Page: '         + str(names_per_page) +     '      ' +
      'Maximum Names per Book: ' + str(max_names_per_book), file=outfile)

print('\nReport any letter whose page count exceeds this size (large letters): ' +
          str(alert_size).rjust(3), file=outfile)
print('\nReport any letter whose page count exceeds this size (near maxpages): ' +
          str(max_limit).rjust(3), file=outfile)
print('\nBook page count at or below which a book is considered to be small  : ' +
          str(small_limit).rjust(3), file=outfile)

print('-' * 83, file=outfile)

print('The summary (multi-precinct) reports are at the end of the output file after all', file=outfile)
print('the individual precinct reports.\n', file=outfile)

print('In both the PACKED and UNPACKED sections of the individual precinct reports, a "-"', file=outfile)
print('is used to denote a missing letter.  For the first pollbook, any "-"s before the', file=outfile)
print('first non-missing letter (usually "A") are trimmed.\n', file=outfile)

print('In the PACKED section of the individual precinct reports, there is a tag in front', file=outfile)
print('of the first letter assigned to each pollbook (except the first book).  The tags', file=outfile)
print('indicate why that pollbook was created.  Only the "bad letter splits" are included', file=outfile)
print('in the split count because the other two do not result in partial letters being', file=outfile)
print('mixed with other letters in the preceding book.\n', file=outfile)

print('  ">" = "bad letter split".  This split letter is in a book with another letter.', file=outfile)
print('  "!" = happenstance.  Pages for prior letters summed to an exact multiple of maxpages.', file=outfile)
print('  "*" = "big" letter split.  The preceding book was maxpages pages of this same letter.\n', file=outfile)

print('In contrast, for the "unpacked" mixrem=no method, the split count is the number of', file=outfile)
print('splits of "big" letters, i.e, the number of additional books produced by those', file=outfile)
print('splits; there are no splits other than those done for "big" letters.  Such', file=outfile)
print('"unpacked" splits never result in partial letters mixing with other letters in any', file=outfile)
print('books and can be considered as "good letter splits" vs. the "bad letter splits"', file=outfile)
print('with the "packed" method or "unpacked" mixrem=yes methods where mixing always', file=outfile)
print('occurs.\n', file=outfile)

print('For the "unpacked" mixrem=yes method, the "bad letter splits" are flagged with a', file=outfile)
print('"@" and are the only things counted in the "unpacked" splits count.  To repeat,', file=outfile)
print('they always result in mixing other letters with the remaining pages of a "big"', file=outfile)
print('letter split in its remainder book.  They are all "bad letter splits".', file=outfile)

print('_' * 83, file=outfile)


#===================  PROCESS LINES FROM INPUT FILE  ===================

# As input this procedure uses a Comma Separated Values (CSV) file
# derived from a WCBOE Excel file.  In the CSV file each precinct has a
# section containing a number of lines, one line per last-name letter of
# precinct voters.  Some letters may be missing if there are no voters
# in the precinct with names beginning with that letter.  Each line has
# these values in this order separated by commas:

# 1) precinct number, e.g., 01-01
# 2) last-name initial letter, A thru Z
# 3) number of voters whose last name starts with this letter
# 4) number of voters rounded up to the nearest multiple of a specified
#       maximum number of names per page  (this was 5 in 2019 pollbooks)
# 5) number of pages required for this rounded-up number of voters

# The procedure goes thru the file, isolates the lines for each
# precinct, and collects the letter data for that precinct.  It then
# calls a function to process the data.

for line in infile:
   linex = line.strip()
   if len(linex) == 0: continue     # Skip blank lines
   if linex[0] == '*': continue     # Allows for comment lines
   if linex[0:3] == '!@#':          # Acts as a stopper line for testing
      finish_precinct()             # Close out the final precinct
      append_multi_reports()        # Output the end-of-file reports
      if mixrem == True: mixrem_text = 'MIXREM=YES'
      else:              mixrem_text = 'MIXREM=NO'
      sys.exit('>>>>> ' + str(gt_precincts) + ' precincts' +
                         '   INPUT=' + in_fileid +
                         '   OUTPUT=' + out_fileid +
                         '   MAXPAGES=' + str(max_pages_per_book) +
                         '   ' + mixrem_text
               + '\n\n!!!!!!!!!! End of Test (stopper line reached)')

   semicolon = linex.find(';')      # Remove comment if present
   if semicolon != -1: linex = linex[0:semicolon]

   linex = linex.strip()
   linex = linex.split(',')         # Get CSV filelds

   # Skip over header line at top of CSV (from Excel) file.
   # Caution if format ever changes.
   if len(linex) > 1 and linex[1] == 'LastINI': continue

   # Skip over grand total line at bottom of file.
   # Caution if format ever changes, e.g., reducing number of commas.
   if len(linex) > 5 and linex[0] == '' and linex[1] == '': continue

   if len(linex) != 5:                 # Make sure there are 5 fields
      print('\n',line)
      sys.exit('ERROR =====> Line found with more or less than five fields.')

   line_precinct = linex[0].strip()    # Get and clean the 5 line values
   line_lastINI  = linex[1].strip()    #
   voters        = linex[2].strip()    #
   ceiling       = linex[3].strip()    #
   pages         = linex[4].strip()    #

   if (len(line_precinct) == 0 or      # Ensure all 5 are not null
       len(line_lastINI)  == 0 or      #
       len(voters)        == 0 or      #
       len(ceiling)       == 0 or      #
       len(pages)         == 0):       #
      print('\n',line)
      sys.exit('ERROR =====> Line contained an empty field.')

   if line_lastINI.isalpha() == False or len(line_lastINI) > 1:
      print('\n',line.strip(), '<-- second field')
      sys.exit('ERROR ====> "' + line_lastINI + ' is not a valid A thru Z single letter.')

   line_lastINI = line_lastINI.upper() # Upper needed for later ord use

   if voters.isdigit() == False:
      print('\n',line.strip(), '<-- third field')
      sys.exit('ERROR ====> "' + voters +  '" is not a non-negative integer.')

   if ceiling.isdigit() == False:
      print('\n',line.strip(), '<-- fourth field')
      sys.exit('ERROR ====> "' + ceiling + '" is not a non-negative integer.')

   if pages.isdigit() == False:
      print('\n',line.strip(), '<-- fifth field')
      sys.exit('ERROR ====> "' + pages +   '" is not a non-negative integer.')


#======  FIND END OF PRECINCT SECTION AND PROCESS THIS PRECINCT  =======

   # Special setup to get the first precinct into the loop.
   if old_precinct == 'first': old_precinct = line_precinct

   # Now the line is known to be clean and can be processed for data.
   # First check if the line is in the same precinct as the one before.
   if line_precinct == old_precinct:            # Still in same precinct
      handle_letter(line_lastINI, pages, old_precinct)  # Process letter

   else:                                        # Precinct has changed
   # Must finish processing the old (current) precinct before starting
   # to process the new one.  More in comment block below.
      finish_precinct()             # Finish processing the old precinct
      handle_letter(line_lastINI, pages, line_precinct) # Process letter

# The finish_precinct function just called completes processing the data
# for the current precinct (preparing the precinct report, etc.), then
# resets variables to switch over to the new precinct and returns.
# Processing of the new precinct (in line_precinct) can now proceed,
# starting with handling the first letter (in line_lastINI with pages
# holding its frequency) whose processing had been postponed while the
# previous precinct had been finished.

# Have reached end of this long "for suite".  Now go back up to the
# "for" to get and process the next line in the file.


#=================  END OF FILE REACHED -- FINISH UP  ==================

# Finish processing the final precinct.
finish_precinct()

# Append reports that consolidate data for all the precincts.
append_multi_reports()

# Close all files.
infile.close()
outfile.close()

if mixrem == True: mixrem_text = 'MIXREM=YES'
else:              mixrem_text = 'MIXREM=NO'

sys.exit('>>>>> ' + str(gt_precincts) + ' precincts' +
                   '   INPUT=' + in_fileid +
                   '   OUTPUT=' + out_fileid +
                   '   MAXPAGES=' + str(max_pages_per_book) +
                   '   ' + mixrem_text + '\n')

#***********************************************************************

# Function call tree

#    main -> handle_letter()

#         -> finish_precinct() -> process_precinct_data()

#         -> append_multi_reports() -> output_precinct_exceptions()
